# -*- coding: utf-8 -*-
"""
Created on Mon May 16 14:07:21 2016

@author: lliu
"""

"""
This log table is optimized for 16-bit unsigned integers on ADI Blackfin
platform.

The table numbers are in format of 3.13, where the three most significant 
bits represent the integer part of a log value and other bits represent 
the fractional part of the log value.

Define log(0) = 0.0

"""
INT_BITS = 3
FRACT_BITS = 16 - INT_BITS
MASK = 0xffff - (0xffff>>(FRACT_BITS)<<(FRACT_BITS))
 


import numpy as np
import math


def generate_log_table():
    
    table = np.zeros((353,1),dtype = 'float')

#     log(0) = 0.0
    num = 1;
    table[0] = 0.0

#    
    for hi in xrange(0,32):
        table[num] = (np.log10((hi<<8) + 63) + np.log10((hi<<8) + 64))/2.0
        num += 1
        table[num] = (np.log10((hi<<8) + 127) + np.log10((hi<<8) + 128))/2.0
        num += 1
        table[num] = (np.log10((hi<<8) + 191) + np.log10((hi<<8) + 192))/2.0
        num += 1
        table[num] = np.log10((hi+1)<<8)
        num += 1
    
    for hi in xrange(32,256):
        table[num] = np.log10((hi+1)<<8)
        num +=1
        
    return table
    
def verify_log_table(table):
    diff = np.zeros((256,256),dtype = 'float')
    diff[0,0] = 0.0
    ######################################################
    # rows 0 - 7
    ######################################################
    
    for hi in xrange(32):
        for lo in xrange(64):
            
            if (hi==0 and lo == 0):
                continue
            

            log1 = table[hi*4]
            log2 = table[hi*4+1]
            e_log = (log2-log1)/64.0*lo + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)
            
        
        for lo in xrange(64,128):
            log1 = table[hi*4+1]
            log2 = table[hi*4+2]
            e_log = (log2-log1)/64.0*(lo-64) + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)        
        
        
        for lo in xrange(128,192):
            log1 = table[hi*4+2]
            log2 = table[hi*4+3]
            e_log = (log2-log1)/64.0*(lo-128) + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)    
        
        
        for lo in xrange(192,256):
            log1 = table[hi*4+3]
            log2 = table[hi*4+4]
            e_log = (log2-log1)/64.0*(lo-192) + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)
    
    ######################################################
    # other rows
    ######################################################
    for hi in xrange(32,256):
        for lo in xrange(256):
            log1 = table[(hi-32) + 128]
            log2 = table[(hi-32) + 1 + 128]
            e_log = (log2-log1)/256.0*lo + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)

    return diff


def generate_log_table1():
    
    table = np.zeros((321,1),dtype = 'float')

#     log(0) = 0.0
    num = 1;
    table[0] = 0.0

#    
    for hi in xrange(0,16):
        table[num] = (np.log10((hi<<8) + 63) + np.log10((hi<<8) + 64))/2.0
        num += 1
        table[num] = (np.log10((hi<<8) + 127) + np.log10((hi<<8) + 128))/2.0
        num += 1
        table[num] = (np.log10((hi<<8) + 191) + np.log10((hi<<8) + 192))/2.0
        num += 1
        table[num] = np.log10((hi+1)<<8)
        num += 1
    
    for hi in xrange(16,32):
        table[num] = (np.log10((hi<<8) + 127) + np.log10((hi<<8) + 128))/2.0
        num += 1
        table[num] = np.log10((hi+1)<<8)
        num += 1
    
    for hi in xrange(32,256):
        table[num] = np.log10((hi+1)<<8)
        num +=1
        
    return table
    
    
def verify_log_table1(table):
    diff = np.zeros((256,256),dtype = 'float')
    diff[0,0] = 0.0
    ######################################################
    # rows 0 - 7
    ######################################################
    
    for hi in xrange(16):
        for lo in xrange(64):
            
            if (hi==0 and lo == 0):
                continue
            

            log1 = table[hi*4]
            log2 = table[hi*4+1]
            e_log = (log2-log1)/64.0*lo + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)
            
        
        for lo in xrange(64,128):
            log1 = table[hi*4+1]
            log2 = table[hi*4+2]
            e_log = (log2-log1)/64.0*(lo-64) + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)        
        
        
        for lo in xrange(128,192):
            log1 = table[hi*4+2]
            log2 = table[hi*4+3]
            e_log = (log2-log1)/64.0*(lo-128) + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)    
        
        
        for lo in xrange(192,256):
            log1 = table[hi*4+3]
            log2 = table[hi*4+4]
            e_log = (log2-log1)/64.0*(lo-192) + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)
            
    ######################################################
    # rows 8 - 15
    ######################################################
    
    for hi in xrange(16,32):
        for lo in xrange(128):
            log1 = table[(hi-16)*2 + 64]
            log2 = table[(hi-16)*2 + 1 + 64]
            e_log = (log2-log1)/128.0*lo + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)
            
        
        for lo in xrange(128,256):
            log1 = table[(hi-16)*2 + 1 + 64]
            log2 = table[(hi-16)*2 + 2 + 64]
            e_log = (log2-log1)/128.0*(lo-128) + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)
    
    ######################################################
    # other rows
    ######################################################
    for hi in xrange(32,256):
        for lo in xrange(256):
            log1 = table[(hi-32) + 96]
            log2 = table[(hi-32) + 1 + 96]
            e_log = (log2-log1)/256.0*lo + log1
            t_log = np.log10((hi<<8)+lo)
            diff[hi,lo] = abs(t_log-e_log)

    return diff
    
    
def verify_int16_table(table_i,table_f):
    diff = np.zeros(table_i.shape,dtype = 'float')
    
    for ii in xrange(table_i.size):

        tmp = table_i[ii]
        diff[ii] = (tmp>>FRACT_BITS) + float((tmp&MASK))/(2**FRACT_BITS) - table_f[ii]
    
    return diff
    
    
    
def convert_to_int16_table(table):
    __table = np.zeros(table.shape,dtype = 'uint32')
    
    for ii in xrange(table.size):
        i_part = int(math.floor(table[ii]))
        f_part = table[ii] - i_part

        tmp = ((i_part)*(2**FRACT_BITS)) + int(2**FRACT_BITS*f_part)
        __table[ii] = tmp
#        print hex(tmp)
    
    return __table
    
    
def log_lookup(num,table):
    
    if(num==0 or num==1):
        return table[0]
    
    hi = num>>8
    lo = num - (hi<<8)
    
    if(hi<32):
        bits = 6
        unit = lo>>bits
        offset = 4*hi + unit

    else:
        bits = 8
        unit = lo>>bits
        offset = hi -32 + 128

    e_val_i = (((table[offset+1]-table[offset])*(lo-unit*(1<<bits)))>>bits) + table[offset]

    
    return e_val_i
    
def log_lookup1(num,table):
    
    if(num==0 or num==1):
        return table[0]
    
    hi = num>>8
    lo = num - (hi<<8)
    
    if(hi<16):
        bits = 6
        unit = lo>>bits
        offset = hi*4 + unit
        
    elif (hi<32):
        bits = 7
        unit = lo>>bits
        offset = (hi-16)*2 + unit + 64

    else:
        bits = 8
        unit = lo>>bits
        offset = hi-32 + 96

    e_val_i = (((table[offset+1]-table[offset])*(lo-unit*(1<<bits)))>>bits) + table[offset]
    
    return e_val_i

    
def table_test(nums,table):
    
    rows,cols = nums.shape
    
    for row in xrange(rows):
        for col in xrange(cols):
            num = nums[row,col]
        
            e_val_i = log_lookup(num,table)
            
            e_val = (e_val_i>>FRACT_BITS) + (e_val_i & MASK)/2**float(FRACT_BITS)    
            t_val = np.log10(num)
            
            print num, e_val, t_val, abs(e_val-t_val)
    
    
    
def random_test(table):
    
    num_cases = 10
    test_case = np.zeros((num_cases,1),dtype='int')
    for ii in xrange(num_cases):
        test_case[ii] = np.random.randint(0,2**16+1)
    
    table_test(np.array(test_case),table)
    
    
    
    
if __name__ == "__main__":

    table_f = generate_log_table()
#    diff = verify_log_table(__table)
    
    table_i = convert_to_int16_table(table_f)
    
    random_test(table_i)
